/* START_LICENSE_HEADER

Copyright (C) 2000 Martin Piper, original design and program code
Copyright (C) 2001-2005 Replica Software

This program file is copyright (C) Replica Software and can only be used under license.
For more information visit: http://www.replicanet.com/
Or email: info@replicanet.com

END_LICENSE_HEADER */
#ifndef __DATABASE_H__
#define __DATABASE_H__
#include "RNPlatform/Inc/DLLExportAPI.h"

namespace RNReplicaNet
{

typedef int t_HashFromDataFunc(void *pData);
typedef int t_HashFromKeyFunc(void *pKey);
typedef bool t_CompareDataWithKeyFunc(void *pData,void *pKey);

// This should always be 32 to use the full 32 bits of a 32 bit integer as a bit array
const int kSubNodes = 32;

class DatabaseNode;
struct DatabaseLinkNode;
class DatabaseLinkNodePool;

class REPNETEXPORTAPI Database
{
friend class DatabaseNode;
public:
	enum Direction
	{
		kForward=0,
		kBackward,
		kUndefined
	};

	Database();
	virtual ~Database();

	void *AddItem(void *data);
	void *AddItemHead(void *data);
	void *FindItem(void *keyData);
	bool RemoveItem(void *data);
	bool RemoveItem(void);

	void BeginIterate(void);
	void EndIterate(void);
	void *Iterate(void);

	void SetHashFromDataFunc(t_HashFromDataFunc *pfunc);
	void SetHashFromKeyFunc(t_HashFromKeyFunc *pfunc);
	void SetCompareDataWithKeyFunc(t_CompareDataWithKeyFunc *pfunc);

	void SetEnableFastFind(bool enable = false);
	bool GetEnableFastFind(void)
	{
		return mEnableFastFind;
	}

	int Tidy(void);

private:
	bool mEnableFastFind;
	Direction mDirection;
	bool mUsed;

	DatabaseNode *MatchNode(int objectHash,bool create);
	void SemiFreeDatabaseLinkNode (DatabaseLinkNode *theLinkNode);

	t_HashFromDataFunc *mHashFromDataFunc;
	t_HashFromKeyFunc *mHashFromKeyFunc;
	t_CompareDataWithKeyFunc *mCompareDataWithKeyFunc;

	// The start of the root node TREE, the mParent should always be null for example
	DatabaseNode *mTreeDatabaseNodeRoot;
	// The start of the link list of allocated root nodes, handy for stepping through when freeing for example
	// Depending on the order of allocation the mParent might not be null
	DatabaseNode *mLinkDatabaseNodeRoot;

	// This pointer is used in the Tidy() routine that goes through the list looking for nodes to tidy
	// To avoid too much time being used up it only does a certain number of nodes per Tody() call
	DatabaseNode *mDatabaseNodeTidyIterator;

	// The iterator values used in the iterator
	DatabaseLinkNode *mIterator;

	// The backup previous values for the iterator
	DatabaseLinkNode *mPrevIterator;

	// The last find puts the entries in here so a subsequent remove (data*) is quicker
	DatabaseNode *mFindCache;
	int mFindCacheCounter;

	// The actual list head and tail used for the iterate LinkNode iterate
	// A totally empty list should have null for both
	DatabaseLinkNode *mLinkNodeRoot;
	DatabaseLinkNode *mLinkNodeTail;

	// The head list pointers of cached pools to avoid too many small allocs that std::list uses
	// A totally empty list should have null pointers for both
	DatabaseLinkNodePool *mPoolRootNode;
	DatabaseLinkNodePool *mPoolFreeRootNode;

	// The number of nodes not immediatly deleted and that contain no entries...
	int mNumNodesWithZeroEntries;
};

class REPNETEXPORTAPI DatabaseNode
{
friend class Database;
public:
	DatabaseNode(DatabaseNode *parent,Database *theDatabase);
	virtual ~DatabaseNode();

protected:
	void AddItem(void *data,DatabaseLinkNode *theLinkNode);

	int mHash;
	DatabaseNode *mParent;
	DatabaseNode *mGreaterThan;
	DatabaseNode *mLessThan;
	DatabaseNode *mSame;

	DatabaseNode *mLinkNext;
	DatabaseNode *mLinkPrev;

	Database *mDatabase;

	unsigned int mBitMask;
	void *mData[kSubNodes];
	DatabaseLinkNode *mTheLinkNode[kSubNodes];
};

struct REPNETEXPORTAPI DatabaseLinkNode
{
	DatabaseLinkNode *mNext;
	DatabaseLinkNode *mPrev;
	void *mData;

	// The pools of link list entries
	DatabaseLinkNodePool *mOwnerPool;
	int mOwnerEntry;

	// The values for the database lookup tree
	DatabaseNode *mDatabaseNode;
	int mDatabaseEntry;
};

class REPNETEXPORTAPI DatabaseLinkNodePool
{
public:
	DatabaseLinkNodePool();
	virtual ~DatabaseLinkNodePool();

	DatabaseLinkNodePool *mNext;
	DatabaseLinkNodePool *mPrev;

	DatabaseLinkNodePool *mNextFree;
	DatabaseLinkNodePool *mPrevFree;

	unsigned int mBitMask;
	DatabaseLinkNode mNodes[kSubNodes];
};

} // namespace RNReplicaNet


#endif
