/* START_LICENSE_HEADER

Copyright (C) 2000 Martin Piper, original design and program code
Copyright (C) 2001-2005 Replica Software

This program file is copyright (C) Replica Software and can only be used under license.
For more information visit: http://www.replicanet.com/
Or email: info@replicanet.com

END_LICENSE_HEADER */
#ifndef __MEMORYTRACKING_H__
#define __MEMORYTRACKING_H__
#include "RNPlatform/Inc/DLLExportAPI.h"

// This file should be included first of all in all C/CPP files to ensure the memory hooks are used.

// These standard includes are pulled in automatically here to enable a clean build without warnings to be done using MSVC.
// Otherwise the MS compiler starts complaining about dll linkage with the allocation routines.
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

// Check for C++ being used
#ifdef __cplusplus
namespace RNReplicaNet
{
}

extern "C"
{
#endif
#include <stddef.h>
typedef size_t tReplicaNet_Size;

typedef void * tReplicaNet_AllocateFunc(tReplicaNet_Size);
typedef void * tReplicaNet_ReAllocateFunc(void *, tReplicaNet_Size);
typedef void * tReplicaNet_ClearAndAllocateFunc(tReplicaNet_Size, tReplicaNet_Size);
typedef void tReplicaNet_FreeFunc(void *);

/**
 * This function allows a function to be set for allocation inside all ReplicaNet libraries.<br>
 * If null is passed in then the default C runtime library function is used.
 */
extern REPNETEXPORTAPI void SetReplicaNetAllocateFunction(tReplicaNet_AllocateFunc *allocateFunction);
/**
 * This function allows a function to be set for re-allocation inside all ReplicaNet libraries.<br>
 * If null is passed in then the default C runtime library function is used.
 */
extern REPNETEXPORTAPI void SetReplicaNetReAllocateFunction(tReplicaNet_ReAllocateFunc *reAllocateFunction);
/**
 * This function allows a function to be set for a clearing allocation inside all ReplicaNet libraries.<br>
 * If null is passed in then the default C runtime library function is used.
 */
extern REPNETEXPORTAPI void SetReplicaNetClearAndAllocateFunction(tReplicaNet_ClearAndAllocateFunc *clearAndAllocateFunction);
/**
 * This function allows a function to be set for freeing inside all ReplicaNet libraries.<br>
 * If null is passed in then the default C runtime library function is used.
 */
extern REPNETEXPORTAPI void SetReplicaNetFreeFunction(tReplicaNet_FreeFunc *freeFunction);

/**
 * Allocates a block of memory.
 * \param size the size in bytes
 * \return the point to the block of memory allocated
 */
extern REPNETEXPORTAPI void *ReplicaNet_AllocateFunction(tReplicaNet_Size size);

/**
 * Allocates or extends a block of memory
 * \param memblock the pointer to the block of memory to extend, can be null
 * \param size the size in bytes
 * \return the point to the block of memory allocated
 */
extern REPNETEXPORTAPI void *ReplicaNet_ReAllocateFunction(void *memblock, tReplicaNet_Size size);

/**
 * Allocates a block of memory.and clears the memory with zero
 * \param num the number of blocks to clear
 * \param size the size of each block in bytes
 * \return the point to the block of memory allocated
 */
extern REPNETEXPORTAPI void *ReplicaNet_ClearAndAllocateFunc(tReplicaNet_Size num, tReplicaNet_Size size);

/**
 * Free a block of memory
 * \param memblock the pointer the the memory
 */
extern REPNETEXPORTAPI void ReplicaNet_FreeFunc(void *memblock);

// Memory total generation

/**
 * Enables memory tracking total generation.<br>
 * Once enabled the total generation should not be disabled unless all memory blocks are freed.
 */
extern REPNETEXPORTAPI void ReplicaNet_MemoryTotalsEnable(void);

/**
 * Disables memory tracking total generation.<br>
 * The total generation should not be disabled unless all memory blocks are freed.
 */
extern REPNETEXPORTAPI void ReplicaNet_MemoryTotalsDisable(void);

/**
 * State detection of the total generator.
 * \return 1 is returned if the total generation is enabled, 0 if disabled. The default is to have total generation disabled.
 */
extern REPNETEXPORTAPI int ReplicaNet_MemoryTotalsState(void);

/**
 * Queries the totals generated so far.
 * \param blocks a pointer to an integer that will contain the number of allocated blocks.
 * \param allocated a pointer to an integer that will contain the number of allocated memory blocks.
 */
extern REPNETEXPORTAPI void ReplicaNet_MemoryTotalsQuery(int *blocks,int *allocated);

/**
 * Queries the peak totals generated so far.
 * \param blocks a pointer to an integer that will contain the number of allocated blocks.
 * \param allocated a pointer to an integer that will contain the number of allocated memory blocks.
 * \param reset Set this to 1 to indicate the peak values should be reset after being queried. 0 Means do not reset the values.
 */
extern REPNETEXPORTAPI void ReplicaNet_MemoryTotalsPeakQuery(int *blocks,int *allocated,int reset);


// Internal function.
extern REPNETEXPORTAPI void ReplicaNet_MemoryDumpAllocatedBlocks(void);
// Internal function.
extern REPNETEXPORTAPI void ReplicaNet_MemoryBreakOnAllocatedBlocks(const int numBlocks,const int *theBlocks);
// Make sure the std calls are redirected to our own functions
#if defined(_WIN32) || defined(_PS2)
#ifndef _MEMORYTRACKING_CPP_
#ifndef REPLICANET_NOMEMORYTRACKING
#define malloc ReplicaNet_AllocateFunction
#define realloc ReplicaNet_ReAllocateFunction
#define calloc ReplicaNet_ClearAndAllocateFunc
#define free ReplicaNet_FreeFunc
#endif
#endif
#endif

#ifdef __cplusplus
};
#endif

#ifdef __cplusplus


#if defined(_WIN32)
#ifndef REPLICANET_NOMEMORYTRACKING
// Claim the new and delete operators for our own purposes
inline void * __cdecl operator new(size_t size)
{
	return ReplicaNet_AllocateFunction(size);
};

inline void __cdecl operator delete(void *p)
{
	ReplicaNet_FreeFunc(p);
};
#endif
#endif


#if defined(_PS2)
#ifndef REPLICANET_NOMEMORYTRACKING
// Claim the new and delete operators for our own purposes
inline void * operator new(size_t size)
{
	return ReplicaNet_AllocateFunction(size);
};

inline void operator delete(void *p)
{
	ReplicaNet_FreeFunc(p);
};
#endif
#endif


#endif

#endif
