/* START_LICENSE_HEADER

Copyright (C) 2000 Martin Piper, original design and program code
Copyright (C) 2001-2005 Replica Software

This program file is copyright (C) Replica Software and can only be used under license.
For more information visit: http://www.replicanet.com/
Or email: info@replicanet.com

END_LICENSE_HEADER */
#ifndef __DATABASEMANAGER_H__
#define __DATABASEMANAGER_H__
#include "RNPlatform/Inc/DLLExportAPI.h"
#include "RNPlatform/Inc/Database.h"

namespace RNReplicaNet
{

/**
 * DatabaseManager is a template class for handling classes in an organised and quick database
 */
template <class T> class DatabaseManager
{
public:
	/**
	 * The ctor performs some basic initialisation
	 */
	DatabaseManager()
	{
	}

	/**
	 * The dtor makes sure everything is tidy
	 */
	virtual ~DatabaseManager()
	{
	}

	/**
	 * Adds an item pointer to the list. This does not allocate an item as this is left to the user to do
	 * \param item the item pointer to add
	 */
	T *AddItem(T *item)
	{
		mDatabase.AddItem(item);
		return item;
	}

	T *AddItemHead(T *item)
	{
		mDatabase.AddItemHead(item);
		return item;
	}

	/**
	 * Removes an item pointer from the list but does not free the item as this is left to the user to do
	 * \param item the item pointer to remove
	 */
	void RemoveItem(T *item)
	{
		mDatabase.RemoveItem(item);
	}

	/**
	 * Removes the current iterated item but does not free the item as this is left to the user to do
	 */
	void RemoveItem(void)
	{
		mDatabase.RemoveItem();
	}

	/**
	 * Starts the iteration of the items in the list from the start and works forwards
	 */
	void BeginIterate(void)
	{
		mDatabase.BeginIterate();
	}

	/**
	 * Starts the iteration of the items in the list from the end and works backwards
	 */
	void EndIterate(void)
	{
		mDatabase.EndIterate();
	}

	/**
	 * Iterates through the items in the list
	 * \return returns the next item pointer in the list or returns NULL for the end of the list
	 */
	T *Iterate(void)
	{
		return (T *)mDatabase.Iterate();
	}

	/**
	 * This finds an item using the t_CompareDataWithKeyFunc as a callback.
	 * If the fast find algorithm is used then the t_HashFromKeyFunc is also used
	 * \param keyData a pointer to key data to search for
	 */
	T *FindItem(void *keyData)
	{
		return (T *)mDatabase.FindItem(keyData);
	}

	int Tidy(void)
	{
		return mDatabase.Tidy();
	}

	/**
	 * A speed enhancement for the database is to turn on the fast find algorithm. This algorithm uses several data hashing functions defined by the user that are then used to index the data.
	 * \param enable set this to be true to enable the fast indexing algorithm. The default is false
	 */
	void SetEnableFastFind(bool enable = false)
	{
		mDatabase.SetEnableFastFind(enable);
	}

	void SetHashFromDataFunc(t_HashFromDataFunc *pfunc)
	{
		mDatabase.SetHashFromDataFunc(pfunc);
	}

	void SetHashFromKeyFunc(t_HashFromKeyFunc *pfunc)
	{
		mDatabase.SetHashFromKeyFunc(pfunc);
	}

	void SetCompareDataWithKeyFunc(t_CompareDataWithKeyFunc *pfunc)
	{
		mDatabase.SetCompareDataWithKeyFunc(pfunc);
	}



private:
	Database mDatabase;
};

} // namespace RNReplicaNet


#endif
