/* START_LICENSE_HEADER

Copyright (C) 2000 Martin Piper, original design and program code
Copyright (C) 2001-2005 Replica Software

This program file is copyright (C) Replica Software and can only be used under license.
For more information visit: http://www.replicanet.com/
Or email: info@replicanet.com

END_LICENSE_HEADER */
#ifndef __DATABLOCK_PREDICT_FLOAT_H__
#define __DATABLOCK_PREDICT_FLOAT_H__
#include "RNPlatform/Inc/DLLExportAPI.h"

#include "RNReplicaNet/Inc/DataBlock.h"
#include <list>

namespace RNReplicaNet
{

/**
 * This DataBlock derived class tries to perform data prediction on float variable types<br>
 * This class operates on three levels.<br>
 * 1) It constantly checks the data value and notes down changes based on what it thinks are continuity changes using the extrapolation model for the master side.<br>
 * 2) Each {value,delta and time} update sent to a replica object is stored and using this information the extrapolation is checked against what the current value is doing. If the extrapolation is too far off based on the input tollerance values then the master copy of the extrapolation data is used to calculate a new more accurate prediction model.<br>
 * 3) The replica version of this class receives {value,delta and time} updates and extrapolates these.<br>
 */
class REPLICANETDLL_API DataBlock_Predict_Float : public DataBlock
{
public:
	/**
	 * The ctor performs some basic initialisation
	 */
	DataBlock_Predict_Float();

	/**
	 * The dtor makes sure everything is tidy
	 */
	virtual ~DataBlock_Predict_Float();

	virtual void Poll(DataBlockReason *reason);

	virtual void ParseMessage(MessageHelper *message);

	virtual void GiveDeltaHint(void *pdata,int dataLen);

	virtual void GiveDeltaHint(const float delta);

	virtual void ContinuityBreak(unsigned char breakTypes);

	bool IsAttached(void *data);

	/**
	 * This registers the memory area with this data block type and returns a new pointer
	 * \param data the start of the memory to check
	 * \return Returns a pointer that can be used to attach to a ReplicaObject
	 */
	static DataBlock_Predict_Float *Register(float *data);

	/**
	 * This registers the memory area with this data block type. This is used when the memory is in datablocks that are allocated as part of a bigger class.
	 * \param data the start of the memory to check
	 */
	void RegisterInClassPointer(float *data)
	{
		mData = data;
	}

	/**
	 * This sets the minimum error that is tolerated for this variable type
	 * \param minerror the minimum error to tolerate
	 */
	void SetMinError(float minerror = 0.1f);

	/**
	 * This sets the maximum error that is tolerated for this variable type
	 * \param maxerror the maximum error to tolerate
	 */
	void SetMaxError(float maxerror = 1.0f);

	/**
	 * Sets the minimum distance start point for the minimum error to take effect
	 * \param the distance
	 */
	void SetMinDistance(float distance = 0.1f);

	/**
	 * Sets the maximum distance start point for the maximum error to take effect
	 * \param the distance
	 */
	void SetMaxDistance(float distance = 1.0f);

	/**
	 * This sets the interpolation filter factor to use when extrapolating this variable type<br>
	 * 0 < interpolation <= 1.0f<br>
	 * 1.0f Is very hard interpolation and 0.1f is very soft. 0.1f Is the default.
	 * \param interpolation the interpolation to use
	 */
	void SetInterpolationFilter(float interpolation = 0.1f);

	/**
	 * For distance based interpolation this makes the calculation based on a quadratic formula. The default is linear.
	 */
	void SetQuadratic(void)
	{
		mIsQuadratic = true;
	}

	/**
	 * For distance based interpolation this makes the calculation based on a linear formula. The default is linear.
	 */
	void SetLinear(void)
	{
		mIsQuadratic = false;
	}

private:
	float *mData;
	bool mContinuityBreak;
	bool mIsDistanceBased;
	float mMinDistance;
	float mMaxDistance;
	float mMinError;
	bool mIsQuadratic;
	bool mGotData;
	bool mImplementTeleport;
	bool mDoPrep;
	bool mCacheValid;



	float mTestData;

	float mMaxError;
	float mInterpolationFilter;


	struct SessionBlock_Predict_Float
	{
		SessionBlock_Predict_Float(int sessid);

		virtual ~SessionBlock_Predict_Float();

		int mSessionID;
		bool mContinuityBreak;
		float mTestData;
		float mDelta;

		float mSample;
		float mSampleTime;

		bool mImplementTeleport;

		// Forced updates
		// This works off local time, not network time
		float mLastUpdateSentLocalTime;
		int mNumForcedUpdates;
	};

	std::list<SessionBlock_Predict_Float *> mPerSessionData;

	struct ReplicaData
	{
		bool mFirstTime;

		float mDelta;
		float mSample;
		float mSampleTime;
	};

	ReplicaData mReplicaData;

	struct MasterData
	{
		float mDelta;
		float mSample;
		float mSampleTime;
		float mSampleHistory[2];
		float mSampleHistoryTime[2];			// This is the network time and is used as values to send off the any replicas
		float mSampleHistoryTimeLocal[2];		// This is the local time and is only used to calculate deltas
		bool mGotDeltaHint;
		float mTheDeltaHint;
	};
	MasterData mMasterData;

	std::list<SessionBlock_Predict_Float *>::iterator mCacheEntry;

	bool mDoneMasterPoll;
};

} // namespace RNReplicaNet

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_VAR macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_VAR(x)	\
	{\
		RNReplicaNet::DataBlock_Predict_Float *datablock = 0;	\
		datablock = RNReplicaNet::DataBlock_Predict_Float::Register(&(basepoint->x));	\
		_RO_DO_SETDATABLOCKVARIABLENAME(x);
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_VAR macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMAXERROR macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMAXERROR(x)	\
		datablock->SetMaxError(x);
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMAXERROR macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMINERROR macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMINERROR(x)	\
		datablock->SetMinError(x);
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMINERROR macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETERROR macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETERROR(x)	\
		datablock->SetMinError(x);	\
		datablock->SetMaxError(x);
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETERROR macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMINDISTANCE macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMINDISTANCE(x)	\
		datablock->SetMinDistance(x);
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMINDISTANCE macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMAXDISTANCE macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMAXDISTANCE(x)	\
		datablock->SetMaxDistance(x);
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETMAXDISTANCE macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETINTERPOLATIONFILTER macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETINTERPOLATIONFILTER(x)	\
		datablock->SetInterpolationFilter(x);
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETINTERPOLATIONFILTER macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETLINEAR macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETLINEAR()	\
		datablock->SetLinear();
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETLINEAR macro ends here */

/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETQUADRATIC macro starts here */
#define _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETQUADRATIC()	\
		datablock->SetQuadratic();
/* _RO_DO_REGISTERBLOCK_PREDICT_FLOAT_SETQUADRATIC macro ends here */

#endif
