/* START_LICENSE_HEADER

Copyright (C) 2000 Martin Piper, original design and program code
Copyright (C) 2001-2005 Replica Software

This program file is copyright (C) Replica Software and can only be used under license.
For more information visit: http://www.replicanet.com/
Or email: info@replicanet.com

END_LICENSE_HEADER */
#ifndef __REPLICAOBJECT_H__
#define __REPLICAOBJECT_H__
#include "RNPlatform/Inc/DLLExportAPI.h"

#include <vector>
#include <list>
#include <set>

#include "RNPlatform/Inc/ThreadClass.h"

namespace RNReplicaNet
{

class ReplicaNet;
class ReplicaNetPrivate;
class DataBlock;
struct s_RO_RegistryBlock;

const float kReplicaObject_InfiniteDistance = -1.0f;	/**< The constant for an infinite distance between two objects. */
const int kReplicaObjectUnknownUniqueID = -1;

/**
 * A ReplicaObject is the base class that all replica objects use.<br>
 * For any ReplicaObject derived class that is intended to be used in a network session an instance of ReplicaNet must be created first.
 * However if a ReplicaObject derived class is not intended to be used in a network session there is no requirement to allocate an instance of ReplicaNet.
 * For example, if an application is started without ReplicaNet networking then instances of ReplicaObject will do nothing and use a nominal amount of memory.<br>
 * Each ReplicaObject class member variable propagated with a DataBlock can have a different packet update type (Unreliable, Certain or Reliable) to the ReplicaObject.
 * When a ReplicaObject is propagated to a session all class member variables that need to send an initial update will temporarily use the update type of the ReplicaObject.
 * This means a ReplicaObject propagated with the Certain update and with a class member variable normally propagated using Unreliable will have the initial variable propagetd using Certain instead.
 * This is to guarantee the initial object state is propagated along with the object propagation packet.
 */
class REPLICANETDLL_API ReplicaObject : private MutexClass
{
public:
	/**
	 * The ctor performs some basic initialisation
	 */
	ReplicaObject();

	/**
	 * The dtor makes sure everything is tidy. If this object has been published the dtor also calls ReplicaNet::RemoveObject(this) to remove the references ReplicaNet holds about this object.
	 */
	virtual ~ReplicaObject();

	/**
	 * Tests the status of the object to see if it is a master object
	 * \return Returns true if this object is a replica (copy) from the network
	 */
	bool IsReplica(void) const;

	/**
	 * Tests the status of the object to see if it is a master object
	 * \return Returns true if this object is a master copy for the network
	 */
	bool IsMaster(void) const;

	/**
	 * Publish this object on to the network. This uses ReplicaNet::AddObject(this) to add a reference of this object to ReplicaNet.
	 */
	void Publish(void);

	/**
	 * Gets the class id that is set
	 * \return the class id for this class
	 */
	virtual const int GetClassID(void) = 0;

	/**
	 * Sets the registered flag to the bool value
	 * \param flag the value to set the flag to, true is default
	 */
	virtual void SetRegistered(bool flag = true) = 0;

	/**
	 * Returns the registered flag status of this class
	 * \return the registered state of this class
	 */
	virtual bool GetRegistered(void) = 0;

	/**
	 * Returns the unique id of the object local to the session. To identify an object for the entire network session you need to use GetUniqueID() and GetSessionID().
	 * \return the unique id
	 */
	int GetUniqueID(void);

	/**
	 * Returns the session id of the object which is global for the whole network session
	 * \return the session id
	 */
	int GetSessionID(void);

	/**
	 * Gets the object's bound ReplicaNet class
	 * \return the ReplicaNet pointer that is bound to this object
	 */
	ReplicaNet *GetBoundReplicaNet(void);

	/**
	 * Use this virtual callback to approve or deny that this object is fault tolerant and can change owner when a session leaves.
	 * \return To approve return true else return false. The default is false
	 */
	virtual bool ApproveFault(void);

	/**
	 * Use this virtual callback to approve or deny that this object can migrate when a migrate request is made.
	 * \return To approve return true else return false. The default is false
	 */
	virtual bool ApproveMigration(void);

	/**
	 * This callback is used whenever an object changes owner
	 */
	virtual void OwnerChange(void);

	/**
	 * This tries to change the ownership on a ReplicaObject to another session.
	 * This only succeeds if the ReplicaObject is a Master object.
	 * This only succeeds if the ApproveMigration() callback returns true for the object.
	 * \param sessionID the target sessionID for this object to migrate to
	 */
	void GiveOwnership(int sessionID);

	/**
	 * This requests a change of the ownership on a ReplicaObject to be the calling session.
	 * This only succeeds if the ReplicaObject is a Replica.
	 * This only succeeds if the ApproveMigration() callback returns true for the object.
	 */
	void RequestOwnership(void);

	/**
	 * The callback for calculating the distance to an object. This can (and should) be replaced by the application's object types.
	 * \param object The destination object to use, if this parameter is null then kReplicaObject_InfiniteDistance should be returned.
	 * \return the distance to the object. The default return value is kReplicaObject_InfiniteDistance meaning that the distance between objects could not be calculated.
	 */
	virtual float CalculateDistanceToObject(ReplicaObject *object);

	/**
	 * Gets the distance to another supplied ReplicaObject. This function applies a caching mechanism to avoid too many unwanted calculations
	 * \param object The destination object to use. If the object pointer is null then kReplicaObject_InfiniteDistance is returned.
	 * \return the distance to the object or kReplicaObject_InfiniteDistance meaning the distance between two objects couldn't be calculated.
	 */
	float GetDistanceToObject(ReplicaObject *object);

	/**
	 * This is a useful virtual call that can be used to set an opaque pointer to classes or data in a ReplicaObject or the derived class
	 * \param data the pointer to the opaque data
	 */
	virtual void SetOpaquePointer(void *data);

	/**
	 * This is a useful virtual call that can be used to obtain opaque pointers to classes or data in a ReplicaObject or the derived class
	 * \param data the pointer to some optional data that can be passed in. This defaults to null
	 * \return the opaque data pointer
	 */
	virtual void *GetOpaquePointer(void *data = 0);

	/**
	 * Given a pointer to memory this functions finds a data block attached to that memory area
	 * \param pdata the pointer to the memory area
	 * \return the DataBlock pointer or NULL if no data block is attached
	 */
	DataBlock *FindDataBlock(void *pdata);

	/**
	 * This gives a delta hint for a particular variable that has a float type
	 * \param variable the variable to give a hint for
	 * \param delta the delta hint
	 */
	void GiveDeltaHint(float &variable,float delta);

	/**
	 * This is used to set a continuity break for the data set and provides useful information to the extrapolation algorithm
	 * \param variable the variable to give a hint for
	 * \param breakTypes The break types to use for this break. For example : DataBlock::ContinuityBreakTypes::kSuddenChange | DataBlock::ContinuityBreakTypes::kTeleport
	 */
	void ContinuityBreak(float &variable,unsigned char breakTypes);

	/**
	 * This sets this object to use the 'reliable' packet type when transmitting information about this object to other sessions.<br>
	 * By default the object uses the reliable method.<br>
	 * Each datablock then uses the type of update assigned to itself.
	 */
	void UpdateSetReliable(void);

	/**
	 * This sets this object to use the 'certain' packet type when transmitting information about this object to other sessions.<br>
	 * By default the object uses the reliable method.<br>
	 * Each datablock then uses the type of update assigned to itself.
	 */
	void UpdateSetCertain(void);

	/**
	 * This sets this object to use the 'normal' packet type when transmitting information about this object to other sessions.<br>
	 * By default the object uses the reliable method.<br>
	 * Each datablock then uses the type of update assigned to itself.
	 */
	void UpdateSetUnreliable(void);

	/**
	 * Tests the update type for this object.
	 * \return true if this object uses the 'reliable' packet type.
	 */
	bool UpdateIsReliable(void);

	/**
	 * Tests the update type for this object.
	 * \return true if this object uses the 'certain' packet type.
	 */
	bool UpdateIsCertain(void);

	/**
	 * Tests the update type for this object.
	 * \return true if this object uses the 'certain' packet type.
	 */
	bool UpdateIsUnreliable(void);

	/**
	 * Enables or disables load balancing for this ReplicaObject.
	 * \param enable set this to true to enable load balancing. The default is false
	 */
	void SetLoadBalancing(const bool enable = false);

	/**
	 * This sets the default value that the load balancing uses. Setting the default value in the object constructor to the same value that is set for the load balancing will result in a smaller data packet being sent.
	 * \param defaultValue the value that is the default value. The default is false to match the parameter to SetLoadBalancing()
	 */
	void SetLoadBalancingDefault(const bool defaultValue = false);

	/**
	 * Gets the status of the load balancing flag
	 * \return true if load balancing is enabled
	 */
	bool GetLoadBalancing(void);

	/**
	 * Sets the object load unit score..
	 * The definition of "load" units is left to the user. The default return value is 0.0f
	 * \param load A "load" score for this object.
	 */
	void SetLoadScore(const float load = 0.0f);

	/**
	 * This sets the default value that the load balancing score uses. Setting the default value in the object constructor to the same value that is set for the load balancing will result in a smaller data packet being sent when an object is published.
	 * When setting a default value the same value must be set on the master and all replica objects otherwise the different sessions will assume different default values. If a replica is unable to know that other replicas will set the same default value then not using the function will ensure that the correct value is propagated.
	 * \param defaultLoad the value that is the default value. The default is 0.0f to match the parameter to SetLoadScore()
	 */
	void SetLoadScoreDefault(const float defaultLoad = 0.0f);

	/**
	 * Gets the object to report a score based on its "load" for the master session.
	 * The definition of "load" units is left to the user. The default return value is 0.0f
	 * \return A "load" score for this object.
	 */
	float GetLoadScore(void);

	/**
	 * This virtual method callback is called by ReplicaNet when a class is created and the initial data from the data blocks has been filled in. The user can choose to override this function if needed. The base class method does not need to be called.
	 */
	virtual void PostObjectCreate(void);

	/**
	 * Allows the user to expand the datablock mechanism with custom registering of datablocks if needed. It is required that the same DataBlocks are registered in the same order for all master and replica objects.
	 * These datablocks are registered after the internal ReplicaObject, ROL derived class datablocks and auto-added datablocks. Any datablocks added by this function should be added using RegisterDataBlock() and not AddAutoRegisterDataBlock()
	 * Do not call this function directly, this function is only called by ReplicaNet.
	 */
	virtual void UserRegisterDataBlocks(void);

	/**
	 * This function allows DataBlocks to automatically register when a ReplicaObject is created.  It is required that the same DataBlocks are registered in the same order for all master and replica objects.
	 * \param dataBlock the pointer to the DataBlock to register
	 */
	void AddAutoRegisterDataBlock(DataBlock *const dataBlock);

	/**
	 * This function allows an object to state which sessionID should be used for sending messages to other sessions.
	 * By default the master object is used to send DataBlock and message updates.
	 * \param sessionID the session to be used for sending message updates, the default is kReplicaObjectUnknownUniqueID which signifies that the master sends messages
	 */
	void SetMessageReflectionID(const int sessionID = kReplicaObjectUnknownUniqueID);

	/**
	 * This function gets the sessionID used to reflect message updates
	 * \return The sessionID that isused to reflect messages, kReplicaObjectUnknownUniqueID means that the master object sends message updates
	 */
	int GetMessageReflectionID(void) const;

	/**
	 * This functions sets the distance that an object has to be within for it to be propagated relative to a session observer.
	 * If a session has an observer and the CalculateDistanceToObject() result is greater than the value specified then the session will not receive object updates.
	 * If a session has no registered observer then the object will always be propagated to that session. It is therefore a good idea to publish and set the observer for a session before joining a ReplicaNet session so that the joining session does not receive objects that will then quickly get removed again.
	 * The default value for a ReplicaObject is to have this distance set to kReplicaObject_InfiniteDistance, which means that the object will always be propogated to a session.
	 * This function can only be used if the ReplicaObject is a master and SetSessionPropagationFilter() is not activated. This distance can be changed at runtime.
	 * To ensure this functionality is enabled the first use of this function must be before the object is published using the Publish() method.
	 * If the ReplicaObject owner changes, notified through the OwnerChange() callback, then SetPropagateToSessionDistance() must be called to enable this functionality on the new master.
	 * \param distance the distance to set, kReplicaObject_InfiniteDistance is the default value
	 */
	void SetPropagateToSessionDistance(const float distance = kReplicaObject_InfiniteDistance);

	/**
	 * This sets the default value that the propagate distance uses. Setting the default value in the object constructor to the same value that is set for the distance will result in a smaller data packet being sent when an object is published.
	 * When setting a default value the same value must be set on the master and all replica objects otherwise the different sessions will assume different default values. If a replica is unable to know that other replicas will set the same default value then not using the function will ensure that the correct value is propagated.
	 * \param distance the value that is the default value. The default is kReplicaObject_InfiniteDistance to match the parameter to SetPropagateToSessionDistance()
	 */
	void SetPropagateToSessionDistanceDefault(const float distance = kReplicaObject_InfiniteDistance);

	/**
	 * Allows the distance value to be read for this ReplicaObject set by SetPropagateToSessionDistance().
	 * This function can be used on a master or replica ReplicaObject
	 * \return the distance value
	 */
	float GetPropagateToSessionDistance(void) const;

	/**
	 * This function sets the minimum recalculation delay that this object will use when it calculates distance based propagation.<br>
	 * If the object owner changes the delay time will revert back to the default value. The object will, if enabled, automatically recalculate the list of sessions to propagate to.
	 * \param delay The recalculation delay in seconds, the default is 0.1 seconds. Values less than 0 are ignored.
	 */
	void SetPropagateToSessionDistanceRecalculationDelay(const float delay = 0.1f);

	/**
	 * This static function can be defined in a derived class to enable the allocation of a ReplicaObjet derived class to be implemented. The default implementation is to use new.
	 * \return A pointer to an object that is derived from ReplicaObject.
	 */
	static ReplicaObject *AllocateForReplicaNet(void);

	/**
	 * This static function can be defined in a derived class to enable the de-allocation of a ReplicaObjet derived class to be implemented. The default implementation is to use delete. Care must be taken to use ReplicaNet::RemoveObject(this) properly to remove references held by ReplicaNet if this function is user implemented.
	 * \param object A pointer to an object that is derived from ReplicaObject.
	 */
	static void DeleteForReplicaNet(ReplicaObject *object);


	/**
	 * This allows the session propagation filter mechanism to be enabled or disabled for a master ReplicaObject.
	 * When enabled the ReplicaObject will ignore all joining sessions and the user application is then responsible for supplying the session IDs that this object will propagate to. This mechanism effectively turns off all ReplicaNet object propagation functions for this object, including the distance based propagation. This mechanism can only be enabled or disabled before the object is published to ReplicaNet.
	 * If the object is set to reflect updates, using SetMessageReflectionID(), then the reflect session ID must also be added to the session propagation list.
	 * If the ReplicaObject owner changes, notified through the OwnerChange() callback, then the previous list of propagated sessions is discarded. SetSessionPropagationFilter() must be called to enable this functionality on the new master and the sessions to propagate to must again be added using SessionPropagationAddSession().
	 * \param enable When true enables this mechanism. The default value is false and this mechanism is disabled.
	 */
	void SetSessionPropagationFilter(const bool enable = false);

	/**
	 * Allows the status of the session propagation filter mechanism to be read.
	 * \param Returns true if the mechanism is enabled
	 */
	bool GetSessionPropagationFilter(void) const;

	/**
	 * Gets the current propagation set.
	 * \param theSet The current set is stored in this reference.
	 */
	void GetSessionPropagationSet(std::set<int> &theSet);

	/**
	 * When the session propagation filter mechanism is enabled this function tells this ReplicaObject to propagate to the supplied session ID. This function will only operate correctly when the session propagation filter mechanism is enabled. The session ID must be a valid ID known to the network session. The session ID from the callback ReplicaNet::JoinerSessionIDPre() can be safely added to objects. If the session ID has already been added to the list (without a corresponding remove) then ReplicaNet will ignore extra additions.
	 * \param sessionID The session ID to propagate to.
	 * \return Returns true if the added sessionID was not in the list. Returns false if the sessionID was in the propagation list.
	 */
	bool SessionPropagationAddSession(const int sessionID);

	/**
	 * When the session propagation filter mechanism is enabled this function tells this ReplicaObject to finish propagating to the supplied session ID. This function will only operate correctly when the session propagation filter mechanism is enabled. The session ID must be a valid ID known to the network session.  The session ID from the callback ReplicaNet::LeaverSessionIDPre() can be safely added to objects. If the session ID has already been removed from the list (without a corresponding add) then ReplicaNet will ignore extra remove requests.
	 * \param sessionID The session ID to finish propagating to.
	 * \return Returns true if the removed sessionID was in the list. Returns false if the sessionID was not in the propagation list.
	 */
	bool SessionPropagationRemoveSession(const int sessionID);

	/**
	 * This virtual callback is triggered by ReplicaNet every time an update packet for a DataBlock is received. See also FindDataBlock()
	 * \param datablock The pointer to the DataBlock that is getting the update.
	 */
	virtual void DataBlockPacketDataReceived(const DataBlock *datablock);

	/**
	 * A pure virtual access method to return the registry block associated with this ReplicaObject class type. This registry block is set when the ReplicaObject derived class is registered with ReplicaNet and the standard interface is implemented by the macro _RO_DO_STD_FRAMEWORK()<br>
	 * For example, to get the registered class name of an unknown "ReplicaObject *theObject" pointer, use the preprocessor to define REPLICANET_VISUALDEBUGGER and then use: theObject->GetRegistryBlock()->GetName().
	 * \return the pointer to the s_RO_RegistryBlock. The registry block contents must not be changed by the user. This function must always return a valid registry block pointer.
	 */
	virtual s_RO_RegistryBlock *GetRegistryBlock(void) const = 0;

protected:
	virtual void _Internal_GetUpdateType(void);
	virtual void _Internal_SetupLoadBalance(void);
	virtual void _Internal_SetupLoadBalanceScore(void);
	virtual void _Internal_SetupLoadBalanceDefault(void);
	virtual void _Internal_SetupLoadBalanceScoreDefault(void);
	virtual void _Internal_SetupOpaquePointer(void);
	virtual void _Internal_SetupPropagateDistance(void);
	virtual void _Internal_SetupPropagateDistanceDefault(void);

	// This sets the internal debugging full name
	void _Internal_SetFullyQualifiedName(const char *name = 0);
	const char *_Internal_GetFullyQualifiedName(void);
	// Debugging emit functions
	void _DebugInternal_EmitClassCreate(const int packetSize,const int toSession);
	void _DebugInternal_EmitClassDelete(const int packetSize,const int toSession);

	/**
	 * Registers a datablock from the derived class to this class. This calls DataBlock::BindReplicaNet(ReplicaObject::GetBoundReplicaNet()), DataBlock::BindReplicaObject(ReplicaObject::this) and DataBlock::SetID(theNextInternalID)
	 * \param The data block to register
	 */
	virtual void RegisterDataBlock(DataBlock *block);

private:
friend class ReplicaNet;			/* ReplicaNet has direct access to ReplicaObject classes */
friend class ReplicaNetPrivate;		/* ReplicaNet has direct access to ReplicaObject classes */
friend struct SessionUnique;		/* ReplicaNet has direct access to ReplicaObject classes */
friend class DataBlock;

// Functions
	// Safe to call multiple times
	virtual void PreRegisterDataBlocks(void);	/**< Pre-registers the data blocks for this ReplicaObject then calls RegisterDataBlocks() */
	// Do not call, except from within PreRegisterDataBlocks()
	virtual void RegisterDataBlocks(void);	/**< Registers the data blocks for this ReplicaObject, virtual because the derived class will want to override the default null behavior of this to register blocks */

	void PollInternalForDistance(const float localTime,ReplicaObject *observer,const int sessionID);

// Variables
	bool mIsMaster;		/**< Flag to say if this class is a master or not */
	int mUniqueID;		/**< Unique ID that is used to reference copies of this object across sessions */
	int mSessionID;		/**< Unique session ID */
	ReplicaNet *mBoundReplicaNet;	/**< The ReplicaNet that is bound to this object */
	bool mBeingDeleted;
protected:
	bool mDataBlocksRegistered;
private:
	bool mFlagForLaterMigration;	/**< This flag is true if the object is flagged for migration when a new master session is chosen */
	void *mOpaquePointer;
	bool mOpaquePointerSet;
	bool mIsReliable;
	bool mIsCertain;
	bool mPostFirstDataCallDone;
	int mTryingToGiveTo;
	bool mObjectHasBeenAdded;	/**< true when the object has been added to ReplicaNet */
	bool mEnableLoadBalancing;
	float mLoadScore;
	bool mEnableLoadBalancingDefault;
	float mLoadScoreDefault;
	int mReflectSessionID;		/**< If not kReplicaNetUnknownUniqueID then this tells ReplicaNet what sessionID is to be used for message refelction */
	float mDistanceToSessionPropagate;
	float mDistanceToSessionPropagateDefault;
	bool mSessionPropagationFilter;

	std::vector<DataBlock *> mAutoRegister;
	std::vector<DataBlock *> mDataBlocks;	/**< The data blocks registered to the derived class */
	std::set<int> mSessionPropagationFilterIDs;
	// These delay slots are for those pending create class or delete class messages that have not been sent yet.
	std::set<int> mSessionPropagationFilterDelaySlotAdd;
	std::set<int> mSessionPropagationFilterDelaySlotRemove;

	// Distance based optimisations
	bool mEnableDistanceBased;
	float mDistanceBasedPollTime;
	float mDistanceBasedLastTime;

	bool mMasterOrReplicaDataBlockPollDone;

	char *mFullyQualifiedName;
	bool mPostObjectCreateDone;
};

} // namespace RNReplicaNet

#endif
